"""ANSI color codes for terminal output"""

RED = "\u001b[1;31m"
GREEN = "\u001b[1;32m"
BLUE = "\u001b[1;34m"
YELLOW = "\u001b[1;33m"
BOLD = "\u001b[1m"
RESET = "\u001b[0m"


def red(content):
    return RED + str(content) + RESET


def green(content):
    return GREEN + str(content) + RESET


def blue(content):
    return BLUE + str(content) + RESET


def yellow(content):
    return YELLOW + str(content) + RESET


def bold(content):
    return BOLD + str(content) + RESET



TRUNC_SUFFIX = " [...]"


class TextPainter:
    def __init__(self, renderer, x_start, y_start, cols, rows, color=None, x_end=None, show_truncate=True):
        self.render = renderer
        self.draw = renderer.draw
        self.font = renderer.body_font
        self.bold_font = renderer.body_bold_font
        self.line_height = renderer.line_height
        self.color = color or renderer.cfg.fg_body

        self.show_truncate = show_truncate

        self.x_start = x_start
        self.y_start = y_start
        self.cols = cols
        self.rows = rows
        self.x_end = x_end or x_start

        self.full = False
        self.cols_used = 0
        self.rows_used = 0
        self.cur_x = x_start
        self.cur_y = y_start
        self.last_line_y = None

    def new_line(self):
        self.rows_used += 1
        if self.rows_used == self.rows:
            # Show truncation indicator if requested and enough space remains
            if self.show_truncate and self.cols_used < self.cols - len(TRUNC_SUFFIX):
                self.write(TRUNC_SUFFIX)

            self.full = True
        else:
            self.cols_used = 0
            self.cur_x = self.x_start
            self.cur_y += self.line_height

    def write(self, text, bold=False, color=None, bg_color=None, return_pos="V"):
        font = self.bold_font if bold else self.font
        color = color or self.color
        last_draw_x = self.cur_x
        last_draw_y = self.cur_y
        tw = 0

        lines = text.split("\n")
        for idx, line in enumerate(lines):
            while not self.full and line:
                # Calculate space and coordinates
                cols_remaining = self.cols - self.cols_used
                y_bottom = self.cur_y - self.line_height
                is_continuation = self.cols_used != 0

                # Draw background if requested
                if bg_color is not None:
                    self.draw.rectangle(((self.cur_x, self.cur_y), (self.x_end, y_bottom)), fill=bg_color)

                draw_seg = line[:cols_remaining]
                # Get text size and center it vertically
                tw, th = self.render.text_size(draw_seg, font=font)
                if is_continuation:
                    center_y = self.last_line_y
                else:
                    center_y = y_bottom + ((self.line_height - th) / 2)
                # Draw text
                self.draw.text((self.cur_x, center_y), draw_seg, font=font, fill=color)
                # Account for drawn text
                self.cur_x += tw
                self.cols_used += len(draw_seg)
                line = line[cols_remaining:]
                self.last_line_y = center_y

                # Advance line if text is remaining
                if line:
                    self.new_line()

            last_draw_x = self.cur_x
            last_draw_y = self.cur_y

            # Advance line if there are more lines
            if idx != len(lines) - 1:
                self.new_line()

        # Return requested position
        if return_pos == "H":
            return last_draw_x - tw / 2, last_draw_y - self.line_height
        elif return_pos == "V":
            return last_draw_x, last_draw_y - self.line_height / 2



from pathlib import Path

from PIL import Image, ImageDraw, ImageFont

from .config import Config
from .gif_encoder import GIFEncoder
from .opencv_encoder import OpenCVEncoder
from .text_format import irepr
from .text_painter import TextPainter
from .webp_encoder import WebPEncoder

WATERMARK = "Generated by vardbg"
SAMPLE_CHARS = "0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ!\"#$%&'()*+,-./:;<=>?@[\\]^_`{|}~ "


class FrameRenderer:
    RED = 0
    GREEN = 1
    BLUE = 2

    def __init__(self, path, config_path):
        # Config
        self.cfg = Config(config_path)
        # Video encoder
        ext = Path(path).suffix.lower()[1:]
        if ext == "mp4":
            self.encoder = OpenCVEncoder(path, "mp4v", self.cfg.fps, self.cfg.w, self.cfg.h)
        elif ext == "gif":
            self.encoder = GIFEncoder(path, self.cfg.fps)
        elif ext == "webp":
            self.encoder = WebPEncoder(path, self.cfg.fps)
        else:
            raise ValueError(f"Unrecognized file extension '{ext}'")
        # Drawing context
        self.draw = None
        # Fonts
        self.body_font = ImageFont.truetype(*self.cfg.font_body)
        self.body_bold_font = ImageFont.truetype(*self.cfg.font_body_bold)
        self.caption_font = ImageFont.truetype(*self.cfg.font_caption)
        self.head_font = ImageFont.truetype(*self.cfg.font_heading)
        self.intro_font = ImageFont.truetype(*self.cfg.font_intro)

        # Whether the watermark has been drawn on this frame
        self._watermark_drawn = False

        # Sizes and positions to be calculated later
        # Code body size
        self.line_height = None
        self.body_cols = None
        self._body_rows = None
        self.body_rows = None
        # Output body start position
        self.out_x = None
        self.out_y = None
        # Output body size
        self.out_cols = None
        self.out_rows = None
        # Variable body start positions
        self.vars_x = None
        self.vars_y = None
        self.ovars_x = None
        self.ovars_y = None
        # Variable body size
        self.vars_cols = None
        self.vars_rows = None
        self.ovars_cols = None
        self.ovars_rows = None

        # Per-frame positions
        self.last_var_x = None
        self.last_var_y = None
        self.ref_var_x = None
        self.ref_var_y = None

        # Current video frame (image)
        self.frame = None

        # Text size cache
        self.text_size_cache = {}

        # Prepare base frame
        self.base_frame = None
        self.prepare_base_frame()

        # Write intro (if necessary)
        if self.cfg.intro_text and self.cfg.intro_time:
            self.write_intro()

    def text_size(self, text, factor=10, **kwargs):
        cache_key = (text, kwargs.get("font", None))

        if cache_key in self.text_size_cache:
            return self.text_size_cache[cache_key]
        else:
            # Multiply string and divide by the factor to get a more precise width
            w, h = self.draw.textsize(text * factor, **kwargs)
            w /= factor

            # Save to cache and return
            sizes = (w, h)
            self.text_size_cache[cache_key] = sizes
            return sizes

    def calc_sizes(self):
        # Calculate text sizes
        w = self.text_size(SAMPLE_CHARS, font=self.body_font)[0] / len(SAMPLE_CHARS)
        hw, hh = self.text_size("A", font=self.head_font)
        _, mh = self.text_size("`^Ag", font=self.body_font)
        _, ch = self.text_size("1p", font=self.caption_font)

        # Code body size
        self.line_height = mh * self.cfg.line_height
        self.body_cols = int((self.cfg.var_x - self.cfg.sect_padding * 2) / w)
        self._body_rows = (self.cfg.out_y - self.cfg.sect_padding * 2 - ch) / self.line_height
        self.body_rows = int(self._body_rows)

        # Output body start position
        self.out_x = self.cfg.sect_padding
        self.out_y = self.cfg.out_y + self.cfg.head_padding * 2 + hh

        # Output body size
        self.out_cols = self.body_cols
        self.out_rows = round((self.cfg.h - self.out_y) / self.line_height)

        # Variable body start positions
        # Top-left X and Y for last variable section
        self.vars_x = self.cfg.var_x + self.cfg.sect_padding
        self.vars_y = self.cfg.head_padding * 2 + hh

        # Columns and rows for last variable section
        self.vars_cols = int((self.cfg.w - self.cfg.var_x - self.cfg.sect_padding * 2) / w)
        self.vars_rows = int((self.cfg.ovar_y - self.cfg.head_padding * 2 - hh) / self.line_height)

        # Top-left X and Y for other variables section
        self.ovars_x = self.vars_x
        self.ovars_y = self.cfg.ovar_y + self.vars_y

        # Columns and rows for other variables section
        self.ovars_cols = self.vars_cols
        ovars_h = self.cfg.h - self.cfg.ovar_y
        self.ovars_rows = int((ovars_h - self.cfg.sect_padding * 2) / self.line_height)

    def get_color(self, col):
        if col == self.RED:
            return self.cfg.red
        elif col == self.GREEN:
            return self.cfg.green
        else:
            return self.cfg.blue

    def draw_text_center(self, x, y, text, font, color):
        w, h = self.text_size(text, font=font)
        self.draw.text((x - w / 2, y - h / 2), text, font=font, fill=color)

    def prepare_base_frame(self):
        # Create new empty frame
        self.new_frame(from_base=False)

        # Draw output section
        # Horizontal divider at 4/5 height
        self.draw.line(((0, self.cfg.out_y), (self.cfg.var_x, self.cfg.out_y)), fill=self.cfg.fg_divider, width=1)
        # Label horizontally centered and padded
        out_center_x = self.cfg.var_x / 2
        out_y = self.cfg.out_y + self.cfg.head_padding
        self.draw_text_center(
            out_center_x, out_y, "Output", self.head_font, self.cfg.fg_heading,
        )

        # Draw variable section
        # Vertical divider at 2/3 width
        self.draw.line(((self.cfg.var_x, 0), (self.cfg.var_x, self.cfg.h)), fill=self.cfg.fg_divider, width=1)
        # Label horizontally centered in the variable section and vertically padded
        var_center_x = self.cfg.var_x + ((self.cfg.w - self.cfg.var_x) / 2)
        self.draw_text_center(var_center_x, self.cfg.head_padding, "Last Variable", self.head_font, self.cfg.fg_heading)

        # Draw other variables section
        # Horizontal divider at 1/3 height
        self.draw.line(
            ((self.cfg.var_x, self.cfg.ovar_y), (self.cfg.w, self.cfg.ovar_y)), fill=self.cfg.fg_divider, width=1
        )
        # Label similar to the first, but in the others section instead
        ovar_label_y = self.cfg.ovar_y + self.cfg.head_padding
        self.draw_text_center(var_center_x, ovar_label_y, "Other Variables", self.head_font, self.cfg.fg_heading)

        # Populate sizes and positions
        self.calc_sizes()

        # Save frame as base and reset current frame
        self.base_frame = self.frame
        self.frame = None

    def new_frame(self, from_base=True):
        # Create image
        if from_base:
            self.frame = self.base_frame.copy()
        else:
            self.frame = Image.new("RGB", (self.cfg.w, self.cfg.h), self.cfg.bg)

        # Create drawing context
        self.draw = ImageDraw.Draw(self.frame)
        # Reset watermark drawn flag
        self._watermark_drawn = False

    def start_frame(self):
        self.new_frame()

    def finish_frame(self, var_state):
        # Bail out if there's no frame to finish
        if self.frame is None:
            return

        # Draw variable state (if available)
        if var_state is not None:
            self.draw_variables(var_state)

        if self.cfg.watermark and not self._watermark_drawn:
            self.draw_watermark()
            self._watermark_drawn = True

        self.encoder.write(self.frame)

    def write_intro(self):
        # Render frame
        self.new_frame(from_base=False)
        x = self.cfg.w / 2
        y = self.cfg.h / 2
        self.draw_text_center(x, y, self.cfg.intro_text, self.intro_font, self.cfg.fg_heading)

        # Repeatedly write frame
        frames = round(self.cfg.intro_time * self.cfg.fps)
        for _ in range(frames):
            self.finish_frame(None)

    def draw_code(self, lines, cur_line):
        cur_idx = cur_line - 1

        # Construct list of (line, highlighted) tuples
        hlines = [(line, i == cur_idx) for i, line in enumerate(lines)]

        # Calculate start and end display indexes with an equivalent number of lines on both sides for context
        ctx_side_lines = (self._body_rows - 1) / 2
        start_idx = round(cur_idx - ctx_side_lines)
        end_idx = round(cur_idx + ctx_side_lines)
        # Accommodate for situations where not enough lines are available at the beginning
        if start_idx < 0:
            start_extra = abs(start_idx)
            end_idx += start_extra
            start_idx = 0
        # Slice selected section
        display_lines = hlines[start_idx:end_idx]

        # Construct painter
        x_start = self.cfg.sect_padding
        y_start = self.cfg.sect_padding + self.line_height
        x_end = self.cfg.var_x - self.cfg.sect_padding
        painter = TextPainter(self, x_start, y_start, self.body_cols, self.body_rows, x_end=x_end, show_truncate=False)

        # Render processed lines
        for i, (line, highlighted) in enumerate(display_lines):
            bg_color = self.cfg.highlight if highlighted else None

            for token, text in line:
                painter.write(text, bg_color=bg_color, **self.cfg.styles[token])

    def draw_output(self, lines):
        lines = lines[-self.out_rows :]
        painter = TextPainter(self, self.out_x, self.out_y, self.out_cols, self.out_rows)
        painter.write("\n".join(lines))

    def draw_exec(self, nr_times, cur, avg, total):
        plural = "" if nr_times == 1 else "s"
        text = f"Line executed {nr_times} time{plural} — current time elapsed: {cur}, average: {avg}, total: {total}"

        _, h = self.text_size(text, font=self.caption_font)
        x = self.cfg.sect_padding
        y = self.cfg.out_y - self.cfg.sect_padding - h
        self.draw.text((x, y), text, font=self.caption_font)

    def draw_last_var(self, state):
        painter = TextPainter(self, self.vars_x, self.vars_y, self.vars_cols, self.vars_rows)

        # Draw variable name
        painter.write(state.name + " ")
        # Draw action with color
        self.last_var_x, self.last_var_y = painter.write(state.action + " ", bold=True, color=state.color)
        painter.new_line()

        # Draw remaining text
        painter.write(state.text)

    def draw_other_vars(self, state):
        painter = TextPainter(self, self.ovars_x, self.ovars_y, self.ovars_cols, self.ovars_rows)

        # Draw text
        for idx, (var, values) in enumerate(state.other_history):
            if values.ignored:
                continue

            if idx > 0:
                painter.write("\n\n")

            painter.write(var.name + ":")

            for v_idx, value in enumerate(values):  # sourcery off
                painter.write("\n    \u2022 ")

                # Reference highlighting for latest value and matching variables only
                if var.name == state.ref and v_idx == len(values) - 1:
                    v_pos = irepr(painter, value.value, state.value, bold=True, color=state.color, return_pos="H")
                    self.ref_var_x, self.ref_var_y = v_pos
                else:
                    irepr(painter, value.value)

    def draw_var_ref(self, state):
        # Calculate X position to route the line on
        # It should be as short as possible while not obscuring any variables or exceeding the scene width
        right_line_x = min(
            max(self.last_var_x, self.ref_var_x) + self.cfg.sect_padding, self.cfg.w - self.cfg.sect_padding / 2
        )

        sw, sh = self.text_size(" ", font=self.body_font)

        # Draw the polyline
        self.draw.line(
            (
                (self.last_var_x, self.last_var_y),
                (right_line_x, self.last_var_y),
                (right_line_x, self.ref_var_y - sh),
                (self.ref_var_x, self.ref_var_y - sh),
                (self.ref_var_x, self.ref_var_y),
            ),
            fill=state.color,
            width=2,
        )

    def draw_variables(self, state):
        self.draw_other_vars(state)
        self.draw_last_var(state)

        if state.ref is not None:
            self.draw_var_ref(state)

    def draw_watermark(self):
        # Get target bottom-right position
        x = self.cfg.w - self.cfg.sect_padding
        y = self.cfg.h - self.cfg.sect_padding

        # Subtract text size to position it properly
        w, h = self.text_size(WATERMARK, font=self.caption_font)
        x -= w
        y -= h

        # Draw text
        self.draw.text((x, y), WATERMARK, fill=self.cfg.fg_watermark, font=self.caption_font)

    def close(self, var_state):
        # Finish final frame
        self.finish_frame(var_state)
        # Close encoder
        self.encoder.stop()



import abc


class Writer(abc.ABC):
    @abc.abstractmethod
    def write_cur_frame(self, frame_info, output):
        pass

    @abc.abstractmethod
    def write_frame_exec(self, frame_info, exec_time, exec_times):
        pass

    @abc.abstractmethod
    def write_add(self, var, val, history, *, action, plural):
        pass

    @abc.abstractmethod
    def write_change(self, var, val_before, val_after, history, *, action):
        pass

    @abc.abstractmethod
    def write_remove(self, var, val, history, *, action):
        pass

    @abc.abstractmethod
    def write_summary(self, var_history, exec_start_time, exec_stop_time, frame_exec_times):
        pass



import collections
import re
import statistics
import textwrap
from pathlib import Path

import pygments
from pygments.lexers.python import PythonLexer
from pygments.token import Token

from ... import render
from ..writer import Writer
from .renderer import FrameRenderer

VarState = collections.namedtuple("VarState", ("name", "color", "action", "value", "ref", "text", "other_history"))


def wrap_text(text, cols, rows=None):
    lines = text.replace("\r", "").split("\n")

    # Wrap text
    wrapped_lines = []
    for line in lines:
        line_wrapped = textwrap.wrap(line, width=cols)
        if len(line_wrapped) == 0:
            wrapped_lines.append("")
        else:
            wrapped_lines += line_wrapped

    # Truncate rows and add indicator if necessary
    if rows is not None and len(wrapped_lines) > rows:
        wrapped_lines = wrapped_lines[:rows]
        wrapped_lines[-1] = wrapped_lines[-1][: cols - 5] + " [...]"

    return wrapped_lines


def list_split(lst, sep):
    new_lst = []
    cur_seg = []

    for elem in lst:
        cur_seg.append(elem)

        if elem == sep:
            new_lst.append(cur_seg)
            cur_seg = []

    return new_lst


class VideoWriter(Writer):
    def __init__(self, path, config_path):
        # File contents
        self.file_cache = {}
        # Current stack frame snapshot (info)
        self.frame_info = None
        # Last variable state
        self.last_var = None
        # Frame renderer
        self.render = FrameRenderer(path, config_path)

    def get_file_lines(self, path):
        if path in self.file_cache:
            return self.file_cache[path]

        lexed = pygments.lex(Path(path).read_text(), PythonLexer())
        lines = list_split(lexed, (Token.Text, "\n"))
        self.file_cache[path] = lines
        return lines

    def write_cur_frame(self, frame_info, output):
        self.frame_info = frame_info
        self.render.finish_frame(self.last_var)
        self.render.start_frame()
        self.render.draw_code(self.get_file_lines(frame_info.file), frame_info.line)
        self.render.draw_output(wrap_text(output, self.render.out_cols))

    def write_frame_exec(self, frame_info, exec_time, exec_times):
        nr_times = len(exec_times)
        avg_time = render.duration_ns(statistics.mean(exec_times))
        total_time = render.duration_ns(sum(exec_times))
        this_time = render.duration_ns(exec_time)

        self.render.draw_exec(nr_times, this_time, avg_time, total_time)

    def _write_action(self, name, val, color, action, fields, history):
        # Render fields
        fields_text = "\n".join(f"{field}: {value}" for field, value in fields.items())

        if history.var_history:
            values = ["\n\nHistory:"]

            for value in history.var_history:
                values.append(repr(value.value))

            history_text = "\n    \u2022 ".join(values)
        else:
            history_text = ""

        # Render and split full text
        text = fields_text + history_text

        # Parse reference comments
        match = re.match(r"^ref (.+)\[(.+)\]$", self.frame_info.comment)
        if match is None:
            ref = None
        else:
            # Extract values
            container = match.group(1)
            key = match.group(2)

            # Make sure both the key matches this variable and the container exists in this scope
            if key == name and any(var.name == container for var, values in history.other_history):
                ref = container
            else:
                ref = None

        # Save state; this is drawn when the frame is finished
        self.last_var = VarState(name, self.render.get_color(color), action, val, ref, text, history.other_history)

    def write_add(self, var, val, history, *, action="added", plural):
        self._write_action(var, val, self.render.GREEN, action, {"Value": repr(val)}, history)

    def write_change(self, var, val_before, val_after, history, *, action="changed"):
        self._write_action(
            var, val_after, self.render.BLUE, action, {"From": repr(val_before), "To": repr(val_after)}, history,
        )

    def write_remove(self, var, val, history, *, action="removed"):
        self._write_action(var, val, self.render.RED, action, {"Last value": repr(val)}, history)

    def write_summary(self, var_history, exec_start_time, exec_stop_time, frame_exec_times):
        # Video doesn't have a summary
        pass

    def close(self):
        self.render.close(self.last_var)



from .gif_encoder import GIFEncoder


class WebPEncoder(GIFEncoder):
    def stop(self):
        # WebP requires an integer duration
        frame_dur = round(1000 / self.fps)
        self.save(frame_dur)



import argparse
import importlib.util
from pathlib import Path

from . import ansi, debugger, test


def parse_args():
    parser = argparse.ArgumentParser(
        description="A simple Python debugger and profiler that can generate animated visualizations of program flow."
    )

    parser.add_argument("-f", "--file", nargs="?", type=str, help="Python file to debug, or JSON result file to read")
    parser.add_argument(
        "-n", "--function", nargs="?", type=str, help="function to run from the given file (if applicable)"
    )
    parser.add_argument(
        "-o",
        "--output-file",
        nargs="?",
        default="debug_results.json",
        help="path to write JSON output file to, default debug_results.json (will be truncated if it already exists and created otherwise)",
    )
    parser.add_argument(
        "-v",
        "--video",
        nargs="?",
        type=str,
        metavar="PATH",
        help="path to write a video representation of the program execution to (MP4, GIF, and WebP formats are supported, depending on file extension)",
    )
    parser.add_argument(
        "-c",
        "--video-config",
        nargs="?",
        default="video.toml",
        metavar="CONFIG",
        help="path to the TOML config for the video output format, default video.toml",
    )
    parser.add_argument(
        "-a", "--args", nargs="*", default=[], metavar="ARGS", help="list of arguments to pass to the running program",
    )
    parser.add_argument(
        "-p",
        "--absolute-paths",
        default=False,
        action="store_true",
        help="use absolute paths instead of relative ones",
    )
    parser.add_argument(
        "-P",
        "--disable-live-profiler",
        default=False,
        action="store_true",
        help="disable live profiler output during execution",
    )

    return parser.parse_args()


def do_debug(args, mod):
    # Get the function here regardless of which path we took above
    func = getattr(mod, args.function, None)
    if func is None:
        # Check how many functions are present in the module
        func_syms = [sym for sym in dir(mod) if callable(getattr(mod, sym))]
        if len(func_syms) == 1:
            # Safe to assume that the user wanted this one if it's the only one
            f_sym = func_syms[0]
            func = getattr(mod, f_sym)
            print(ansi.yellow(f"Unable to find function '{args.function}', falling back to the only one: '{f_sym}'"))
            print()
        else:
            # Ambiguous if multiple, so bail out and let the user choose
            print(ansi.red(f"Unable to find function '{args.function}' and multiple are present; aborting."))
            return 1

    # Call the actual debugger with our parameters
    debugger.debug(
        func,
        args=[args.file or func.__code__.co_filename, *args.args],
        relative_paths=not args.absolute_paths,
        json_out_path=args.output_file,
        video_out_path=args.video,
        video_config=args.video_config,
        live_profiler_output=not args.disable_live_profiler,
    )
    return 0


def do_replay(args):
    debugger.replay(args.file, video_out_path=args.video, video_config=args.video_config)


def main():
    args = parse_args()

    # Module defaults to test
    # Function defaults to test_func if no file specified, otherwise main
    if args.file is None:
        mod = test
        if args.function is None:
            args.function = "test_func"

        # Debug the test function
        return do_debug(args, mod)
    else:
        # Use pathlib to get more info
        file_path = Path(args.file)
        if file_path.suffix == ".json":
            # JSON file means to replay, not debug
            return do_replay(args)
        else:
            if args.function is None:
                args.function = "main"

            # Load file as module and debug
            mod_name = Path(args.file).stem
            spec = importlib.util.spec_from_file_location(mod_name, args.file)
            mod = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(mod)

            return do_debug(args, mod)



import cv2
import numpy as np

from .encoder import Encoder


class OpenCVEncoder(Encoder):
    # noinspection PyUnresolvedReferences
    def __init__(self, path, fourcc, fps, w, h):
        cv_fourcc = cv2.VideoWriter_fourcc(*fourcc)
        self.writer = cv2.VideoWriter(path, cv_fourcc, fps, (w, h))

    def write(self, image):
        # Convert PIL -> Numpy array and RGB -> BGR colors
        # noinspection PyUnresolvedReferences
        cv_img = cv2.cvtColor(np.asarray(image), cv2.COLOR_RGB2BGR)
        # Write data
        self.writer.write(cv_img)

    def stop(self):
        self.writer.release()



from . import ansi


def key_var(var, key):
    return f"{var}[{repr(key)}]"


def val(value):
    return ansi.bold(repr(value))


def duration_ns(t_ns) -> str:
    t_ns = int(t_ns)

    t_us = t_ns / 1000
    t_ms = t_us / 1000
    t_s = t_ms / 1000
    t_m = t_s / 60
    t_h = t_m / 60
    t_d = t_h / 24

    if t_d >= 1:
        rem_h = t_h % 24
        rem_m = t_m % 60
        rem_s = t_s % (24 * 60 * 60) % 60
        return "%dd %dh %dm %ds" % (t_d, rem_h, rem_m, rem_s)
    elif t_h >= 1:
        rem_m = t_m % 60
        rem_s = t_s % (60 * 60) % 60
        return "%dh %dm %ds" % (t_h, rem_m, rem_s)
    elif t_m >= 1:
        rem_s = t_s % 60
        return "%dm %ds" % (t_m, rem_s)
    elif t_s >= 10:
        return "%ds" % t_s
    elif t_ms >= 10:
        return "%d ms" % t_ms
    elif t_us >= 1:
        return "%d μs" % t_us
    else:
        return "%d ns" % t_ns



import statistics
import sys

from .. import ansi, data, render
from .writer import Writer


class ConsoleWriter(Writer):
    def __init__(self, file=None):
        # Output file
        self.file = file or sys.stdout
        # Current line output prefix
        self.cur_line = ""
        # Last stdout output length
        self.stdout_last_len = 0

    def print(self, *args, **kwargs):
        print(*args, **kwargs, file=self.file)

    def write_cur_frame(self, frame_info, output):
        # Construct friendly filename + line number + function string
        file_line = "%s:%-2d" % (frame_info.file, frame_info.line)
        self.cur_line = f"{file_line} ({frame_info.function})"

        # Print new stdout output
        new_output = output[self.stdout_last_len :]
        if new_output:
            self.file.write(new_output)
        self.stdout_last_len = len(output)

    def write_frame_exec(self, frame_info, exec_time, exec_times):
        nr_times = len(exec_times)
        avg_time = render.duration_ns(statistics.mean(exec_times))
        total_time = render.duration_ns(sum(exec_times))
        this_time = render.duration_ns(exec_time)

        self.print(f"{self.cur_line} | exec {nr_times}x (time: {this_time}, avg {avg_time}, total {total_time})")

    def _write_action(self, var, color_func, action, suffix):
        self.print(f"{self.cur_line} | {ansi.bold(var)} {color_func(action)} {suffix}")

    def write_add(self, var, val, history, *, action, plural):
        _plural = "s" if plural else ""
        self._write_action(var, ansi.green, action, f"with value{_plural} {render.val(val)}")

    def write_change(self, var, val_before, val_after, history, *, action):
        self._write_action(
            var, ansi.blue, action, f"from {render.val(val_before)} to {render.val(val_after)}",
        )

    def write_remove(self, var, val, history, *, action):
        self._write_action(var, ansi.red, action, f"(value: {render.val(val)})")

    def _write_var_summary(self, var_history):
        self.print()
        self.print("Variables seen:")

        for var, values in var_history.items():
            # Check whether all the values were numbers
            if all(isinstance(val.value, (int, float)) for val in values):
                # Give a min-max range for numbers
                min_val = min(values, key=data.VarValue.value_getter)
                max_val = max(values, key=data.VarValue.value_getter)

                min_desc = f"{render.val(min_val.value)} ({min_val.file_line})"
                max_desc = f"{render.val(max_val.value)} ({max_val.file_line})"
                values_desc = f" between {min_desc} and {max_desc}"
            else:
                # Give a list of values for other objects
                value_lines = [":"]

                for val in values:
                    value_lines.append(f"{render.val(val.value)} ({val.file_line})")

                # Format list
                values_desc = "\n      - ".join(value_lines)

            definition = f"in {var.function} on {ansi.bold(var.file_line)}"
            self.print(f"  - {ansi.bold(var.name)} {ansi.green('defined')} {definition} with values{values_desc}")

            if values.deleted_line is not None:
                self.print(f"    ({ansi.red('deleted')} on {values.deleted_line})")

    def _write_line_summary(self, frame_exec_times):
        self.print()
        self.print("Lines executed:")

        for frame_info, exec_times in frame_exec_times.items():
            nr_times = len(exec_times)
            avg_time = render.duration_ns(statistics.mean(exec_times))
            total_time = render.duration_ns(sum(exec_times))

            file_line = "%s:%-2d (%s)" % (frame_info.file, frame_info.line, frame_info.function)
            self.print(
                f"{file_line} | {ansi.bold(nr_times)}x, avg {ansi.bold(avg_time)}, total {ansi.bold(total_time)}"
            )

    def _write_time_summary(self, exec_start_time, exec_stop_time):
        self.print()

        exec_time = render.duration_ns(exec_stop_time - exec_start_time)
        self.print(f"Total execution time: {ansi.bold(exec_time)}")

    def write_summary(self, var_history, exec_start_time, exec_stop_time, frame_exec_times):
        self._write_var_summary(var_history)
        self._write_line_summary(frame_exec_times)
        self._write_time_summary(exec_start_time, exec_stop_time)

    def close(self):
        # We print everything live, so there's nothing to close
        pass



import copy
import time

import jsonpickle

from .writer import Writer

NEW_FRAME = "new_frame"
EXECUTE_FRAME = "exec_frame"

ADD_VARIABLE = "add_var"
CHANGE_VARIABLE = "change_var"
REMOVE_VARIABLE = "remove_var"


class JsonWriter(Writer):
    def __init__(self, output_path):
        self.output_path = output_path

        self.data = {"events": []}
        self._step = 0

    def step(self):
        self._step += 1
        return self._step

    def write_event(self, evt_name, **kwargs):
        event = {"step": self.step(), "time": time.time_ns(), "event": evt_name}
        event.update(kwargs)

        self.data["events"].append(event)

    def write_cur_frame(self, frame_info, output):
        self.write_event(NEW_FRAME, frame_info=frame_info, output=output)

    def write_frame_exec(self, frame_info, exec_time, exec_times):
        # exec_times needs to be copied to preserve the *current* state
        self.write_event(EXECUTE_FRAME, frame_info=frame_info, exec_time=exec_time, exec_times=exec_times.copy())

    def write_add(self, var, val, history, *, action, plural):
        self.write_event(
            ADD_VARIABLE, var_name=var, value=val, history=copy.deepcopy(history), action=action, plural=plural,
        )

    def write_change(self, var, val_before, val_after, history, *, action):
        self.write_event(
            CHANGE_VARIABLE,
            var_name=var,
            value_before=val_before,
            value_after=val_after,
            history=copy.deepcopy(history),
            action=action,
        )

    def write_remove(self, var, val, history, *, action):
        self.write_event(REMOVE_VARIABLE, var_name=var, value=val, history=copy.deepcopy(history), action=action)

    def write_summary(self, var_history, exec_start_time, exec_stop_time, frame_exec_times):
        # frame_exec_times is skipped because it can be readily reconstructed during replay
        self.data["var_history"] = list(var_history.items())
        self.data["exec_start_time"] = exec_start_time
        self.data["exec_stop_time"] = exec_stop_time

    def close(self):
        # Write all the collected data out together
        with open(self.output_path, "w+") as f:
            f.write(jsonpickle.dumps(self.data))



from pathlib import Path

import toml
from pygments.formatter import Formatter
from pygments.styles.monokai import MonokaiStyle
from pygments.token import Token

BASE_PATH = Path(__file__).parent / ".." / ".." / ".."
DEFAULT_CFG_PATH = BASE_PATH / "config.toml"


def load_data(path):
    # Attempt to load file
    try:
        data = path.read_text()
    except FileNotFoundError:
        # Fall back to default config if not found
        data = DEFAULT_CFG_PATH.read_text()

    return data


def sub_path(path):
    # Replace first $VARDBG with base path
    if path.startswith("$VARDBG"):
        path = path.replace("$VARDBG", str(BASE_PATH), 1)

    return path


def calc_frac(value, frac):
    num, denom = frac
    return round(value * num / denom)


def parse_hex_color(string):
    if string.startswith("#"):
        string = string[1:]

    r = int(string[0:2], 16)
    g = int(string[2:4], 16)
    b = int(string[4:6], 16)

    return r, g, b, 255


def load_style(style):
    styles = {}
    formatter = Formatter(style=style)

    for token, params in formatter.style:
        color = parse_hex_color(params["color"]) if params["color"] else None
        # Italic and underline styles aren't supported, so just use bold for them
        bold = params["bold"] or params["italic"] or params["underline"]

        # Save style
        styles[token] = {"color": color, "bold": bold}

    return styles


class Config:
    def __init__(self, config_path):
        # Load config
        self.data = load_data(Path(config_path))
        self.config = toml.loads(self.data)

        # Extract values
        general = self.config["general"]
        self.w = general["width"]
        self.h = general["height"]
        self.fps = general["fps"]

        self.intro_text = general["intro_text"]
        self.intro_time = general["intro_time"]
        self.watermark = general["watermark"]

        sizes = self.config["sizes"]
        self.var_x = calc_frac(self.w, sizes["code_width"])
        self.out_y = calc_frac(self.h, sizes["code_height"])
        self.ovar_y = calc_frac(self.h, sizes["last_variable_height"])

        self.head_padding = sizes["heading_padding"]
        self.sect_padding = sizes["section_padding"]

        self.line_height = sizes["line_height"]

        fonts = self.config["fonts"]
        self.font_body = (sub_path(fonts["body"]), fonts["body_size"])
        self.font_body_bold = (sub_path(fonts["body_bold"]), fonts["body_size"])
        self.font_caption = (sub_path(fonts["caption"]), fonts["caption_size"])
        self.font_heading = (sub_path(fonts["heading"]), fonts["heading_size"])
        self.font_intro = (sub_path(fonts["intro"]), fonts["intro_size"])

        style = MonokaiStyle
        self.styles = load_style(style)

        self.bg = parse_hex_color(style.background_color)
        self.highlight = parse_hex_color(style.highlight_color)
        self.fg_divider = self.styles[Token.Generic.Subheading]["color"]
        self.fg_heading = self.styles[Token.Name]["color"]
        self.fg_body = self.styles[Token.Text]["color"]
        self.fg_watermark = self.styles[Token.Comment]["color"]

        self.red = self.styles[Token.Operator]["color"]
        self.green = self.styles[Token.Name.Function]["color"]
        self.blue = self.styles[Token.Keyword]["color"]



from .encoder import Encoder


class GIFEncoder(Encoder):
    def __init__(self, path, fps):
        self.frames = []
        self.path = path
        self.fps = fps

    def write(self, image):
        self.frames.append(image)

    def save(self, frame_dur):
        self.frames[0].save(self.path, save_all=True, append_images=self.frames[1:], duration=frame_dur, loop=0)

    def stop(self):
        frame_dur = 1000 / self.fps
        self.save(frame_dur)



import os.path
import re
from pathlib import Path

_relative_path_cache = {}


def _get_path(orig_path, relative):
    if relative:
        if orig_path in _relative_path_cache:
            return _relative_path_cache[orig_path]
        else:
            rel = os.path.relpath(orig_path)
            _relative_path_cache[orig_path] = rel
            return rel
    else:
        return orig_path


class Variable:
    """Holds information about a variable"""

    def __init__(self, name, frame_info):
        # Basic variable info
        self.name = name

        # Extract info from frame
        self._file = frame_info.file
        self.file_line = frame_info.file_line
        self.function = frame_info.function

    def to_tuple(self):
        # This produces an identifying tuple for hashing and equality comparison.
        # We ignore value, type, and line here because they can change
        return self.name, self._file, self.function

    def __hash__(self):
        return hash(self.to_tuple())

    def __eq__(self, other):
        return self.to_tuple() == other.to_tuple()

    def __ne__(self, other):
        return not (self == other)


class VarValues(list):
    def __init__(self, *values, ignored=False):
        super().__init__(values)
        self.ignored = ignored
        self.deleted_line = None  # file:line


class VarValue:
    """Holds information about a variable value"""

    def __init__(self, value, frame_info):
        self.value = value
        self.file_line = frame_info.file_line

    @staticmethod
    def value_getter(val):
        return val.value


class VarHistory:
    """Holds information about a variable's history and its context"""

    def __init__(self, var, full_history):
        self.var = var
        self.var_history = full_history[var] if var in full_history else ()
        other_history = {v: h for v, h in full_history.items() if v != var and v.function == var.function}
        self.other_history = list(other_history.items())


class FrameInfo:
    """Holds basic information about a stack frame"""

    def __init__(self, frame, *, relative=True, file_cache=None):
        self.function = frame.f_code.co_name
        self.file = _get_path(frame.f_code.co_filename, relative)
        self.line = frame.f_lineno

        self.file_line = f"{self.file}:{self.line}"

        # Extract vardbg comment from line
        line = self._get_line(file_cache)
        match = re.match(r"^.+# vardbg: (.+)$", line)
        self.comment = "" if match is None else match.group(1)

    def _get_line(self, file_cache):
        # Read lines from cache if possible, otherwise read from file and save to cache
        if file_cache and self.file in file_cache:
            lines = file_cache[self.file]
        else:
            lines = Path(self.file).read_text().replace("\r", "").splitlines()
            file_cache[self.file] = lines

        return lines[self.line - 1]

    def to_tuple(self):
        # Produce an identifying tuple for hashing and equality comparison
        return self.file, self.line, self.function

    def __hash__(self):
        return hash(self.to_tuple())

    def __eq__(self, other):
        return self.to_tuple() == other.to_tuple()

    def __ne__(self, other):
        return not (self == other)

    def __lt__(self, other):
        return self.line < other.line



import abc
from typing import TYPE_CHECKING

import dictdiffer

from . import data, render

if TYPE_CHECKING:
    from .debugger import Debugger


class DiffProcessor(abc.ABC):
    def __init__(self: "Debugger"):
        # Full variable + values map
        self.vars = {}

        # File contents cache
        self.file_cache = {}

        # Propagate initialization to other mixins
        super().__init__()

    def _get_history(self, wrapper):
        return data.VarHistory(wrapper, self.vars)

    def process_add(self: "Debugger", chg_name, chg, frame_info, new_locals):
        # If we have a changed variable, elements were added to a list/set/dict
        if chg_name:
            # Get a reference to the container to check its type
            container = new_locals[chg_name]
            # Construct variable wrapper
            wrapper = data.Variable(chg_name, frame_info)

            if not self.vars[wrapper].ignored:
                # chg is a list of tuples with keys (index, key, etc.) and values
                for key, val in chg:
                    if isinstance(container, set):
                        # Move value out of set if there's only 1
                        if len(val) == 1:
                            val = val.pop()

                        # Show it as an extension for sets
                        self.out.write_add(
                            chg_name, val, self._get_history(wrapper), action="extended", plural=isinstance(val, set)
                        )
                    else:
                        # Render it as var[key] for lists, dicts, etc.
                        self.out.write_add(
                            render.key_var(chg_name, key), val, self._get_history(wrapper), action="added", plural=False
                        )

                # Record new value
                self.vars[wrapper].append(data.VarValue(container, frame_info))

        # Otherwise, it's a new variable
        else:
            # chg is a list of tuples with variable names and values
            for name, val in chg:
                wrapper = data.Variable(name, frame_info)
                ignored = frame_info.comment == "ignore"
                if ignored:
                    self.vars[wrapper] = data.VarValues(ignored=True)
                else:
                    self.out.write_add(name, val, self._get_history(wrapper), action="added", plural=False)
                    self.vars[wrapper] = data.VarValues(data.VarValue(val, frame_info))

    def process_change(self: "Debugger", chg_name, chg, frame_info, new_locals):
        before, after = chg

        # If the changed variable is given as a list, a list/set/dict element was changed
        if isinstance(chg_name, list):
            # chg_name is a tuple with the variable name and key
            var_name, key = chg_name
            # Render it as var_name[key]
            chg_name = f"{var_name}[{repr(key)}]"

            # Full changed value
            full_after = new_locals[var_name]
        else:
            var_name = chg_name
            full_after = after

        wrapper = data.Variable(var_name, frame_info)
        if not self.vars[wrapper].ignored:
            self.out.write_change(chg_name, before, after, self._get_history(wrapper), action="changed")
            self.vars[wrapper].append(data.VarValue(full_after, frame_info))

    def process_remove(self: "Debugger", chg_name, chg, frame_info, new_locals):
        # If we have a changed variable, elements were removed from a list/set/dict
        if chg_name:
            # Construct variable wrapper
            wrapper = data.Variable(chg_name, frame_info)

            if not self.vars[wrapper].ignored:
                for key, val in chg:
                    self.out.write_remove(
                        render.key_var(chg_name, key), val, self._get_history(wrapper), action="removed"
                    )

                # Get new container contents and log value
                container = new_locals[chg_name]
                self.vars[wrapper].append(data.VarValue(container, frame_info))

        # Otherwise, a variable was deleted
        else:
            # chg is a list of tuples with variable names and values
            for name, val in chg:
                # Construct variable wrapper
                wrapper = data.Variable(name, frame_info)

                if not self.vars[wrapper].ignored:
                    self.out.write_remove(name, val, self._get_history(wrapper), action="deleted")
                    self.vars[wrapper].deleted_line = frame_info.file_line

    def process_locals_diff(self: "Debugger", diff, frame_info, new_locals):
        for action, chg_var, chg in diff:
            if action == dictdiffer.ADD:
                self.process_add(chg_var, chg, frame_info, new_locals)
            elif action == dictdiffer.CHANGE:
                self.process_change(chg_var, chg, frame_info, new_locals)
            elif action == dictdiffer.REMOVE:
                self.process_remove(chg_var, chg, frame_info, new_locals)

    def finalize_history(self: "Debugger"):
        # Delete ignored variables (implementation detail) from history map
        to_delete = [var for var, values in self.vars.items() if values.ignored]
        for var in to_delete:
            del self.vars[var]



import abc
import copy
import io
import sys
from typing import TYPE_CHECKING

import dictdiffer

from . import data, internal

if TYPE_CHECKING:
    from .debugger import Debugger


ALLOWED_EVENTS = {"call", "line", "return"}
DISALLOWED_FUNC_NAMES = {"<genexpr>", "<listcomp>", "<dictcomp>", "<setcomp>"}


class FrameScope:
    """Scope of one stack frame and its snapshots."""

    def __init__(self):
        # Previous frame and its locals
        self.prev_frame_info = None
        self.prev_event = None
        self.prev_locals = {}
        # New frame's locals
        self.new_locals = {}


class Tracer(abc.ABC):
    def __init__(self: "Debugger"):
        # Function being debugged
        self.func = None
        # Stack with frame scopes
        self.scope_stack = []

        # stdout buffer
        self.stdout_buf = io.StringIO()

        # File content cache
        self.file_cache = {}

        # Propagate initialization to other mixins
        super().__init__()

    def trace_callback(self: "Debugger", frame, event, arg):
        """Frame execution callback"""

        # Ignore irrelevant events, but still attach to the next one
        if event not in ALLOWED_EVENTS:
            return self.trace_callback

        # Ignore our internal functions
        code = frame.f_code
        if code in internal.INTERNAL_FUNC_CODES:
            return

        # Ignore comprehensions and generator expressions
        # (they act strangely and most people wouldn't consider them to be functions)
        if code.co_name in DISALLOWED_FUNC_NAMES:
            return

        # Obtain frame scope
        if event == "call":
            # Create and push a new scope for this stack frame (used for all of its snapshots)
            scope = FrameScope()
            self.scope_stack.append(scope)
        else:
            # Don't use pop since we may need to reuse the scope
            scope = self.scope_stack[-1]

        # The first frame is when function arguments are populated, so it's important
        # Set itself to the previous frame since its line number *is* where function arguments are defined
        frame_info = data.FrameInfo(frame, relative=self.use_relative_paths, file_cache=self.file_cache)
        if scope.prev_frame_info is None:
            scope.prev_frame_info = frame_info
            scope.prev_event = event

        # Only invoke profiler if the last event was a normal line, since calls haven't executed anything yet
        # and returns aren't actually relevant to the code
        should_profile = scope.prev_event == "line"
        if should_profile:
            # Call profiler first to avoid counting the time it takes to copy locals
            self.profile_complete_frame(scope.prev_frame_info)

        # Get new locals and copy them so they don't change on the next frame
        scope.new_locals = copy.deepcopy(frame.f_locals)

        # Render output prefix for this frame
        self.out.write_cur_frame(scope.prev_frame_info, self.stdout_buf.getvalue())

        # Skip profiler for the first frame since it's before any real execution (just the function call)
        if should_profile:
            # Call profiler to print the last frame's execution
            self.profile_print_frame(scope.prev_frame_info)

        # Diff and print changes
        diff = dictdiffer.diff(scope.prev_locals, scope.new_locals)
        self.process_locals_diff(diff, scope.prev_frame_info, scope.new_locals)

        if event == "return":
            # Use pop() to return to the previous frame since the scope is no longer needed
            self.scope_stack.pop()
        else:
            # Update previous frame info in preparation for the next frame
            scope.prev_frame_info = frame_info
            scope.prev_event = event
            scope.prev_locals = scope.new_locals

        # Don't profile returns (performance isn't user code) or calls (nothing's actually executed yet)
        if event == "line":
            self.profile_start_frame()

        # Subscribe to the next frame, if any
        return self.trace_callback

    def run(self: "Debugger", func, *args, **kwargs):
        # Set function
        self.func = func

        # Override system context
        old_args = sys.argv
        if self.args is not None:
            sys.argv = self.args

        real_stdout = sys.stdout
        sys.stdout = self.stdout_buf

        # Run function with trace callback registered
        sys.settrace(self.trace_callback)
        self.profile_start_exec()
        ret = self.func(*args, **kwargs)
        self.profile_end_exec()
        sys.settrace(None)

        # Restore arguments
        sys.argv = old_args
        sys.stdout = real_stdout

        # Finalize variable history
        self.finalize_history()

        # Write summary with collected data
        self.out.write_summary(self.vars, self.exec_start_time, self.exec_stop_time, self.frame_exec_times)
        return ret

    internal.add_funcs(run)



from .writer import Writer


class OutputDelegate(Writer):
    def __init__(self, *writers):
        self.writers = writers

    def write_cur_frame(self, *args, **kwargs):
        for writer in self.writers:
            writer.write_cur_frame(*args, **kwargs)

    def write_frame_exec(self, *args, **kwargs):
        for writer in self.writers:
            writer.write_frame_exec(*args, **kwargs)

    def write_add(self, *args, **kwargs):
        for writer in self.writers:
            writer.write_add(*args, **kwargs)

    def write_change(self, *args, **kwargs):
        for writer in self.writers:
            writer.write_change(*args, **kwargs)

    def write_remove(self, *args, **kwargs):
        for writer in self.writers:
            writer.write_remove(*args, **kwargs)

    def write_summary(self, *args, **kwargs):
        for writer in self.writers:
            writer.write_summary(*args, **kwargs)

    def close(self, *args, **kwargs):
        for writer in self.writers:
            writer.close(*args, **kwargs)



#!/usr/bin/env python3

import math
import sys
import time


# Sample function to get the value of tau (pi * 2)
def get_tau():
    x = math.pi
    x *= 2
    return x


# Example function to debug
def test_func():  # sourcery off
    # Print given arguments
    print("Arguments:", " ".join(sys.argv[1:]))

    # Simple variable assignments
    y = 9
    x = y * 2
    x *= 2
    print("x =", x)

    # Hidden variable
    c = 1  # vardbg: ignore

    # Adding to a list
    # noinspection PyListCreation
    lst = [1]
    lst.append(x)
    lst.append(y)

    # Deleting from a list
    del lst[2], lst[1]
    print("lst =", lst)

    # Set modification
    st = {1}
    st.add(x)
    st.add(y)
    st.update({1, 2, 3, 4, 5})
    st.remove(x)
    print("st =", st)

    # Dict modification
    # noinspection PyDictCreation
    dct = {"a": 1, "b": 2, "c": 3}
    dct["d"] = 4
    del dct["c"]
    print("dct =", dct)

    # Multiple assignments on a single line
    v1, v2 = 0, 1
    v1, v2 = 1, 2

    # Multiple deletions on a single line
    del v1, v2

    # Executing lines multiple times and profiling them
    lst += [6, 7, 12, 14]
    # Variable relationships
    for i in range(len(lst)):  # vardbg: ref lst[i]
        print("i =", i)
        time.sleep(0.01)
        pass  # This is an intentionally very long line that exists solely for the purpose of testing line wrapping functionality in the video output writer's text renderer, ignore and do not wrap this line so that it remains useful for its purpose

    # Get tau from another function
    tau = get_tau()
    print("tau =", tau)

    # To facilitate verifying the debugger's values
    print("\nFinal test values:")
    print("\n".join(f"  {var} = {repr(val)}" for var, val in locals().items()))  # Convoluted to avoid mutating locals



# Set of code objects of internal debugger functions that are called when tracing is active
INTERNAL_FUNC_CODES = set()


def add_funcs(*funcs):
    INTERNAL_FUNC_CODES.update(set(fn.__code__ for fn in funcs))



import collections.abc


def _irepr_seq_iterable(painter, lst, highlight_idx=None, end_symbols=("[", "]"), *args, **kwargs):
    # Initialize state
    painter.write(end_symbols[0])
    elem_pos = None

    for idx, elem in enumerate(lst):
        # Add comma separator if this isn't the first element
        if idx > 0:
            painter.write(", ")

        # If this is our candidate element,
        if idx == highlight_idx:
            # Record the position and pass highlight flags
            elem_pos = painter.write(repr(elem), *args, **kwargs)
        else:
            painter.write(repr(elem))

    # Finalize string
    last_pos = painter.write(end_symbols[1])

    # Mark end as highlighted target if not found
    if elem_pos is None:
        elem_pos = last_pos

    # Return indices
    return elem_pos


def _irepr_dict(painter, dct, highlight_key=None, *args, **kwargs):
    # Initialize state
    painter.write("{")
    elem_pos = None

    for idx, (key, elem) in enumerate(dct.items()):
        # Add comma separator if this isn't the first element
        if idx > 0:
            painter.write(", ")

        # Add key first (no need to record this)
        painter.write(repr(key) + ": ")

        # If this is our candidate element,
        if key == highlight_key:
            # Record the position and pass highlight flags
            elem_pos = painter.write(repr(elem), *args, **kwargs)
        else:
            painter.write(repr(elem))

    # Finalize string
    last_pos = painter.write("}")

    # Mark end as highlighted target if not found
    if elem_pos is None:
        elem_pos = last_pos

    # Return indices
    return elem_pos


def irepr(painter, obj, *args, **kwargs):
    # Compare and invoke container-specific implementations if applicable to the given object
    # We use ABC collections classes to catch all similar classes, e.g. sortedcontainers objects
    if isinstance(obj, collections.abc.Sequence):
        return _irepr_seq_iterable(painter, obj, *args, **kwargs)
    elif isinstance(obj, collections.abc.Mapping):
        return _irepr_dict(painter, obj, *args, **kwargs)
    elif isinstance(obj, collections.abc.Set):
        # Same thing as sequences, just different symbols
        return _irepr_seq_iterable(painter, obj, *args, end_symbols=("{", "}"), **kwargs)
    else:
        # Fall back to a generic implementation with standard repr()
        return painter.write(repr(obj))



import abc
from typing import TYPE_CHECKING

import jsonpickle

from .output.json_writer import ADD_VARIABLE, CHANGE_VARIABLE, EXECUTE_FRAME, NEW_FRAME, REMOVE_VARIABLE

if TYPE_CHECKING:
    from .debugger import Debugger


class Replayer(abc.ABC):
    def __init__(self: "Debugger"):
        # Propagate initialization to other mixins
        super().__init__()

    def replay_events(self: "Debugger", events):
        for event in events:
            evt_type = event["event"]

            if evt_type == NEW_FRAME:
                self.out.write_cur_frame(event["frame_info"], event["output"])
            elif evt_type == EXECUTE_FRAME:
                frame_info = event["frame_info"]
                exec_time = event["exec_time"]

                self.out.write_frame_exec(frame_info, exec_time, event["exec_times"])

                # Replay changes to frame_exec_times
                if frame_info in self.frame_exec_times:
                    self.frame_exec_times[frame_info].append(exec_time)
                else:
                    self.frame_exec_times[frame_info] = [exec_time]
            elif evt_type == ADD_VARIABLE:
                self.out.write_add(
                    event["var_name"], event["value"], event["history"], action=event["action"], plural=event["plural"],
                )
            elif evt_type == CHANGE_VARIABLE:
                self.out.write_change(
                    event["var_name"],
                    event["value_before"],
                    event["value_after"],
                    event["history"],
                    action=event["action"],
                )
            elif evt_type == REMOVE_VARIABLE:
                self.out.write_remove(event["var_name"], event["value"], event["history"], action=event["action"])
            else:
                raise ValueError(f"Unrecognized JSON event '{evt_type}'")

    def replay_summary(self: "Debugger", data):
        self.vars.update(data["var_history"])
        self.out.write_summary(self.vars, data["exec_start_time"], data["exec_stop_time"], self.frame_exec_times)

    def replay(self: "Debugger", json_path):
        with open(json_path, "r") as f:
            data = jsonpickle.loads(f.read())

        self.replay_events(data["events"])
        self.replay_summary(data)



import abc
import platform
import time
from typing import TYPE_CHECKING

import sortedcontainers

from . import internal

if TYPE_CHECKING:
    from .debugger import Debugger


# Helper to call a function from the time module with nanosecond precision if possible, and lower prrcision otherwise
def _time_wrap_ns(sym_base):
    sym_ns = sym_base + "_ns"
    if hasattr(time, sym_ns):
        # Use ns function if possible
        return getattr(time, sym_ns)
    else:
        # Otherwise, create a wrapper to convert it
        # This is necessary on Python < 3.7
        func = getattr(time, sym_base)
        return lambda: int(func() * 1_000_000_000)


# Use performance counter on Windows and process time on others for max precision (ref: PEP-564)
if platform.system() == "Windows":
    get_time_ns = _time_wrap_ns("perf_counter")
else:
    get_time_ns = _time_wrap_ns("process_time")


class Profiler(abc.ABC):
    def __init__(self: "Debugger"):
        # Map of FrameInfo objects and a list of their execution times in ns
        self.frame_exec_times = sortedcontainers.SortedDict()
        # When the last frame started executing
        self.prev_frame_start_time = None

        # Overall start and stop times
        self.exec_start_time = None
        self.exec_stop_time = None

        # Propagate initialization to other mixins
        super().__init__()

    def profile_start_frame(self: "Debugger"):
        self.prev_frame_start_time = get_time_ns()

    def profile_complete_frame(self: "Debugger", prev_frame_info):
        exec_time = get_time_ns() - self.prev_frame_start_time

        if prev_frame_info in self.frame_exec_times:
            self.frame_exec_times[prev_frame_info].append(exec_time)
        else:
            self.frame_exec_times[prev_frame_info] = [exec_time]

    def profile_print_frame(self: "Debugger", prev_frame_info):
        if self.live_profiler_output:
            exec_times = self.frame_exec_times[prev_frame_info]
            self.out.write_frame_exec(prev_frame_info, exec_times[-1], exec_times)

    def profile_start_exec(self: "Debugger"):
        self.exec_start_time = time.time_ns()

    def profile_end_exec(self: "Debugger"):
        self.exec_stop_time = time.time_ns()

    internal.add_funcs(profile_start_exec, profile_end_exec)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['vardbg', 'vardbg.output', 'vardbg.output.video_writer']

package_data = \
{'': ['*']}

install_requires = \
['dictdiffer>=0.8.1,<0.9.0',
 'jsonpickle>=1.2,<2.0',
 'opencv-python>=4.1.2,<5.0.0',
 'pillow>=7.0.0,<8.0.0',
 'pygments>=2.5.2,<3.0.0',
 'sortedcontainers>=2.1.0,<3.0.0',
 'toml>=0.10.0,<0.11.0']

entry_points = \
{'console_scripts': ['pyrobud = vardbg.main:main']}

setup_kwargs = {
    'name': 'vardbg',
    'version': '0.11.4',
    'description': 'A simple Python debugger and profiler that generates animated visualizations of program flow.',
    'long_description': "# vardbg\n\nA simple Python debugger and profiler that generates animated visualizations of program flow.\n\n**Python 3.6** or newer is required due to the use of f-strings.\n\nThis project was created during [Google Code-in](https://codein.withgoogle.com/) 2019 for [CCExtractor Development](https://ccextractor.org/).\n\n## Features\n\n- Tracking the history of each variable and its contents\n- Tracking elements within containers (lists, sets, dicts, etc.)\n- Ignoring specific variables\n- Profiling the execution of each line\n- Summarizing all variables and execution times after execution\n- Passing arguments to debugged programs\n- Exporting execution history in JSON format and replaying (including program output)\n- Creating videos that show program flow, execution times, variables (with relationships), and output\n- Writing videos in MP4, GIF, and WebP formats\n\n## Demo\n\n![Demo Video](https://user-images.githubusercontent.com/7930239/72689524-12691180-3ac7-11ea-9547-861454b1496d.gif)\n\n## Installation\n\nThe latest tagged version can be obtained from PyPI:\n\n```bash\npip install vardbg\n```\n\n## Usage\n\nAll of the program's options are documented in the usage help:\n\n```\n$ vardbg --help\nusage: vardbg [-h] [-f [FILE]] [-n [FUNCTION]] [-o [OUTPUT_FILE]] [-v [PATH]] [-c [CONFIG]]\n              [-a [ARGS [ARGS ...]]] [-p] [-P]\n\nA simple debugger that traces local variable changes, lines, and times.\n\noptional arguments:\n  -h, --help            show this help message and exit\n  -f [FILE], --file [FILE]\n                        Python file to debug, or JSON result file to read\n  -n [FUNCTION], --function [FUNCTION]\n                        function to run from the given file (if applicable)\n  -o [OUTPUT_FILE], --output-file [OUTPUT_FILE]\n                        path to write JSON output file to, default debug_results.json (will be truncated if it\n                        already exists and created otherwise)\n  -v [PATH], --video [PATH]\n                        path to write a video representation of the program execution to (MP4 and GIF formats\n                        are supported, depending on file extension)\n  -c [CONFIG], --video-config [CONFIG]\n                        path to the TOML config for the video output format, default video.toml\n  -a [ARGS [ARGS ...]], --args [ARGS [ARGS ...]]\n                        list of arguments to pass to the running program\n  -p, --absolute-paths  use absolute paths instead of relative ones\n  -P, --disable-live-profiler\n                        disable live profiler output during execution\n```\n\nFor example, this command will debug the function `quick_sort` from the file `sort.py` with the arguments `9 3 5 1` and create a JSON file called `sort1.json`:\n\n```bash\nvardbg -f sort.py -n quick_sort -o sort1.json -a 9 3 5 1\n```\n\nA video can then be generated from the above JSON:\n\n```bash\nvardbg -f sort1.json -v sort_vis.mp4\n```\n\nIt is possible to generate videos live while running the debugged program, but this is discouraged because the overhead of video creation inflates execution times greatly and thus ruins profiler results. However, if profiling is not important to you, it is a valid use case.\n\n## Comments\n\nSpecial comments can be added to lines of code that define variables to control how vardbg handles said variable:\n\n- `# vardbg: ignore` — do not display this variable or track its values\n- `# vardbg: ref lst[i]` — treat variable `i` as the index/key of an element in container `lst` (only shown in videos)\n\n## Contributing\n\nFeel free to contribute to this project! You can add features, fix bugs, or make any other improvements you see fit. We just ask that you follow the [code style guidelines](https://github.com/CCExtractor/vardbg/blob/master/CODE_STYLE.md) to keep the code consistent and coherent. Once your contribution meets the guidelines, [open a pull request](https://github.com/CCExtractor/vardbg/compare) to make things official.\n",
    'author': 'Danny Lin',
    'author_email': 'danny@kdrag0n.dev',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/CCExtractor/vardbg',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



from . import output
from .diff_processor import DiffProcessor
from .profiler import Profiler
from .replayer import Replayer
from .tracer import Tracer


class Debugger(DiffProcessor, Profiler, Replayer, Tracer):
    def __init__(
        self,
        args=None,
        relative_paths=True,
        json_out_path=None,
        video_out_path=None,
        video_config=None,
        live_profiler_output=True,
    ):
        # Arguments to pass to snippet (handled in run())
        self.args = args

        # Whether to use relative paths over absolute ones in output
        self.use_relative_paths = relative_paths

        # Whether to print profiler output live during execution
        self.live_profiler_output = live_profiler_output

        # Output writers
        writers = [output.ConsoleWriter()]
        if json_out_path is not None:
            writers.append(output.JsonWriter(json_out_path))
        if video_out_path is not None:
            writers.append(output.VideoWriter(video_out_path, video_config))
        self.out = output.OutputDelegate(*writers)

        # Initialize mixins
        super().__init__()

    def close(self):
        self.out.close()

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()


def debug(func, *args, **kwargs):
    with Debugger(*args, **kwargs) as dbg:
        dbg.run(func)


def replay(json_path, *args, **kwargs):
    with Debugger(*args, **kwargs) as dbg:
        dbg.replay(json_path)
